import processing.net.*;
import omicronAPI.*;

OmicronAPI omicronManager;
TouchListener touchListener;

ProjectThree applet;
boolean usingWall = false;

MenuState ms;
Menu r;
Fractal f;
WidgetColor wc;

public void init() {
  super.init();
  omicronManager = new OmicronAPI(this);
  omicronManager.setFullscreen(true);
}

void setup() {
  if (usingWall) {
    size(8160, 2304, P3D);
    omicronManager.connectToTracker(7001, 7340 /*28000*/, "crepe.evl.uic.edu");
  }
  else size(displayWidth, displayHeight, P3D);
  touchListener = new TouchListener();
  omicronManager.setTouchListener(touchListener);
  applet = this;

  ms = new MenuState();
  textAlign(LEFT, TOP);
  PFont font = loadFont("LucidaGrande-Bold-60.vlw");
  textFont(font, ms.scale);

  r = new Menu("Root", null);
  r.addMenu("Close");
  r.addMenu("Fractal Color");
  r.addMenu("Fractal Options");
  r.children.get(2).addMenu("Display Complex Grid");
  r.children.get(2).addMenu("Display Orbit Trap");
  r.children.get(2).addMenu("Julia / Mandelbrot Mode");
  r.addMenu("Navigation");
  r.children.get(3).addMenu("Pan");
  r.children.get(3).addMenu("Reset");
  r.children.get(3).addMenu("Zoom In");
  r.children.get(3).addMenu("Zoom Out");

  f = new Fractal();
  wc = new WidgetColor();
  f.shaderRender();
}

void draw() {
  image(f.pg, 0, 0, width, height);
  if (ms.mode.equals("Display Orbit Trap")) f.displayOrbitTrap();
  ms.display();
  if (wc.visible) {
    wc.display();
    if (!wc.conflictingNodeMenu()) r.display(ms.x, ms.y);
  }
  else if (r.visible) r.display(ms.x, ms.y);
  omicronManager.process();
}

void touchDown(int ID, float xPos, float yPos, float xWidth, float yWidth) {
  if (r.visible) {
    Menu selectedMenu = r.pointInMenu(ms.x, ms.y, int(xPos), int(yPos));
    if (selectedMenu != null && !wc.conflictingNodeMenu()) {
      r.hideAll();
      selectedMenu.showMenuPath();
      r.correctPosition(0, 0);
      selectedMenu.doAction();
    }
    else {
      boolean maintainWcMode = false;
      if (!wc.mode.equals("")) {
        maintainWcMode = true;
        wc.pointInNode(int(xPos), int(yPos));

        if (wc.mode.equals("Add") && wc.nodes.size() < 32) {
          String previousMode = wc.mode;
          wc.pointInOption(int(xPos), int(yPos));
          if (wc.mode.equals("")) wc.addNode(int(xPos), 0xff000000|int(random(0xffffff)));
          wc.mode = previousMode;
        }
        else if (wc.mode.equals("Color") && wc.selected != -1) {
          wc.nodes.get(wc.selected).updating = !wc.nodes.get(wc.selected).updating;
          wc.conflictingNodeUpdates();
        }
        else if (wc.mode.equals("Remove") && wc.selected != -1 && wc.nodes.size() > 2) wc.removeNode(wc.selected);
      }

      boolean checkMsModes = true;
      if (wc.visible) {
        String previousMode = wc.mode;
        wc.pointInOption(int(xPos), int(yPos));

        if (maintainWcMode) {
          checkMsModes = false;
          if (wc.mode.equals("")) wc.mode = previousMode;
          else if (wc.mode.equals(previousMode)) wc.mode = "";
          if (!wc.mode.equals("Color")) wc.hideUpdate();
        }
      }

      if (checkMsModes && wc.mode.equals("")) {
        if (ms.mode.equals("Display Orbit Trap")) {
          int maxPwh = max(width, height);
          float delta = f.wh/maxPwh*0.5;
          float xOffset = min((width-height)*delta, 0.0);
          float yOffset = min((height-width)*delta, 0.0);
          f.or = map(xPos, 0.0, width, f.xMin-xOffset, f.xMin+f.wh+xOffset);
          f.oi = map(yPos, 0.0, height, f.yMin+f.wh+yOffset, f.yMin-yOffset);
        }
        else if (ms.mode.equals("Pan")) {
          int maxPwh = max(width, height);
          float delta = f.wh/maxPwh*0.5;
          float offset = f.wh*0.5;
          f.xMin += (xPos/maxPwh)*f.wh-min((width-height)*delta, 0.0)-offset;
          f.yMin += ((height-yPos)/maxPwh)*f.wh-min((height-width)*delta, 0.0)-offset;

          if (f.juliaMode) {
            f.jSet.set("xMin", f.xMin);
            f.jSet.set("yMin", f.yMin);
          }
          else {
            f.mSet.set("xMin", f.xMin);
            f.mSet.set("yMin", f.yMin);
          }
          f.shaderRender();
        }
        else if (ms.mode.equals("Set Constant")) {
          f.jr = map(xPos, 0.0, width, f.xMin, f.xMin+f.wh);
          f.ji = map(yPos, 0.0, height, f.yMin+f.wh, f.yMin);
          f.jSet.set("jr", f.jr);
          f.jSet.set("ji", f.ji);
          f.shaderRender();
        }
        else if (ms.mode.equals("Zoom In")) {
          int maxPwh = max(width, height);
          float delta = f.wh/maxPwh*0.5;
          float offset = f.wh*0.25;
          f.xMin += (xPos/maxPwh)*f.wh-min((width-height)*delta, 0.0)-offset;
          f.yMin += ((height-yPos)/maxPwh)*f.wh-min((height-width)*delta, 0.0)-offset;
          f.wh *= 0.5;

          if (f.juliaMode) {
            f.jSet.set("wh", f.wh);
            f.jSet.set("xMin", f.xMin);
            f.jSet.set("yMin", f.yMin);
          }
          else {
            f.mSet.set("wh", f.wh);
            f.mSet.set("xMin", f.xMin);
            f.mSet.set("yMin", f.yMin);
          }
          f.shaderRender();
        }
        else if ((ms.mode.equals("Zoom Out"))) {
          int maxPwh = max(width, height);
          float delta = f.wh/maxPwh*0.5;
          f.xMin += (xPos/maxPwh)*f.wh-min((width-height)*delta, 0.0)-f.wh;
          f.yMin += ((height-yPos)/maxPwh)*f.wh-min((height-width)*delta, 0.0)-f.wh;
          f.wh *= 2.0;

          if (f.juliaMode) {
            f.jSet.set("wh", f.wh);
            f.jSet.set("xMin", f.xMin);
            f.jSet.set("yMin", f.yMin);
          }
          else {
            f.mSet.set("wh", f.wh);
            f.mSet.set("xMin", f.xMin);
            f.mSet.set("yMin", f.yMin);
          }
          f.shaderRender();
        }
        else {
          ms.x = int(xPos)-(r.w>>1);
          ms.y = int(yPos)-(r.h>>1);
          r.correctPosition(0, 0);
          if (wc.visible) wc.correctPosition();
        }
      }
    }
  }
  else {
    r.visible = true;
    ms.x = int(xPos)-(r.w>>1);
    ms.y = int(yPos)-(r.h>>1);
    r.correctPosition(0, 0);
  }
}

float previousX = -1.0;
void touchMove(int ID, float xPos, float yPos, float xWidth, float yWidth) {
  if (previousX != xPos) {
    if (wc.mode.equals("Color")) {
      boolean updated = false;
      for (int i = 0; i < wc.nodes.size(); i++) {
        if (wc.nodes.get(i).updating) {
          int startLeft = wc.nodes.get(i).x-(ms.scale*3)-ms.halfScale;
          startLeft = max(min(startLeft, width-(ms.scale<<3)-ms.scale)+ms.quarterScale, ms.scale);

          int startRight = startLeft+(ms.scale<<3)+ms.halfScale;
          int startTop = ms.y+r.h+ms.oneAndHalfScale-ms.doubleScale;
          for (int j = 0; j < 3; j++) {
            int currTop = startTop-j*ms.scale;
            if (xPos >= startLeft && xPos < startRight && yPos >= currTop && yPos < currTop+ms.halfScale) {
              int jShift = j<<3;
              int mapPos = int(map(xPos, startLeft+1, startRight-1, 0, 255));
              mapPos = max(min(mapPos, 255), 0)<<jShift;
              wc.nodes.get(i).c = (wc.nodes.get(i).c&(~(0xff<<jShift)))|mapPos;

              updated = true;
              break;
            }
          }
        }
      }
      if (updated) wc.shaderUpdateNodes();
    }
    else if (wc.mode.equals("Move") && wc.selected != -1) {
      int constrainedVal = max(min(int(xPos), width-ms.scale), ms.halfScale);
      wc.nodes.get(wc.selected).x = constrainedVal;
      wc.shaderUpdateNodes();
    }
  }
  previousX = xPos;
}

void touchUp(int ID, float xPos, float yPos, float xWidth, float yWidth) {
  if (wc.mode.equals("Color")) {
    wc.shaderUpdateNodes();
    f.shaderRender();
  }
  if (wc.mode.equals("Move") && wc.selected != -1) {
    wc.sortNodes();
    wc.shaderUpdateNodes();
    f.shaderRender();
  }
  wc.selected = -1;
}

class MenuState {
  int quarterScale, halfScale, scale, oneAndHalfScale, doubleScale;

  boolean showGrid;
  int x, y;
  String mode;

  MenuState() {
    scale = (usingWall) ? 60 : 20;
    quarterScale = scale>>2;
    halfScale = scale>>1;
    oneAndHalfScale = scale+halfScale;
    doubleScale = scale<<1;

    showGrid = false;
    x = halfScale;
    mode = "";
  }

  void display() {
    String state = (r.visible) ? "The menu can be repositioned by touching outside the menu" : "Touch to open menu";
    if (mode.equals("Display Complex Grid")) {
      state = "Displaying real numbers on x-axis and imaginary numbers on y-axis";
    }
    else if (mode.equals("Display Orbit Trap")) {
      state = "Displaying orbit trap of ("+f.or+", "+f.oi+"i), "+f.oIterations+" iterations of possible "+f.maxIterations;
    }
    else if (mode.equals("Fractal Color")) {
      if (wc.mode.equals("")) state = "Touch \"Color\", \"Move\", \"Add\", or \"Remove\" to modify color nodes";
      else if (wc.mode.equals("Color")) state = "Touch a color node to modify its color";
      else if (wc.mode.equals("Move")) state = "Touch and drag a color node to move it";
      else if (wc.mode.equals("Add")) {
        state = (wc.nodes.size() < 32) ? "Touch to add a color node" : "At most 32 color nodes can be added";
      }
      else if (wc.mode.equals("Remove")) {
        state = (wc.nodes.size() > 2) ? "Touch a color node to remove" : "At least 2 color nodes are required";
      }
    }
    else if (mode.equals("Julia / Mandelbrot Mode")) {
      state = (f.juliaMode) ? "Displaying Julia fractal: ("+f.jr+", "+f.ji+"i)" : "Displaying Mandelbrot fractal";
    }
    else if (mode.equals("Pan")) state = "Touch a location to pan";
    else if (mode.equals("Reset")) state = "Complex plane window top-left: ("+f.xMin+", "+f.yMin+"i) and length: "+f.wh;
    else if (mode.equals("Set Constant")) {
      state = "Touch to select c where c = (r, i). Current value of c: ("+f.jr+", "+f.ji+"i)";
    }
    else if (mode.equals("Zoom In")) state = "Touch a location to zoom towards";
    else if (mode.equals("Zoom Out")) state = "Touch a location to zoom away from";

    if (showGrid) {
      fill(0x80202020);
      noStroke();
      rect(0, 0, width, height);

      fill(0xffe0e0e0);
      stroke(0x40e0e0e0);

      line(0, height>>1, width, height>>1);
      line(width>>1, 0, width>>1, height);

      int inc = ms.scale<<3;
      for (int i = (width>>1)+inc; i < width-ms.scale; i += inc) {
        line(i, 0, i, height);
        String mapR = map(i, 0, width, f.xMin, f.xMin+f.wh)+"";
        text(mapR, i-(int(textWidth(mapR))>>1), (height>>1)-ms.halfScale);
      }
      for (int i = (width>>1)-inc; i > ms.scale; i -= inc) {
        line(i, 0, i, height);
        String mapR = map(i, 0, width, f.xMin, f.xMin+f.wh)+"";
        text(mapR, i-(int(textWidth(mapR))>>1), (height>>1)-ms.halfScale);
      }

      inc = ms.scale<<2;
      for (int i = (height>>1)+inc; i < height-ms.scale; i += inc) {
        line(0, i, width, i);
        String mapI = map(i, 0, height, f.yMin+f.wh, f.yMin)+"i";
        text(mapI, (width-int(textWidth(mapI)))>>1, i-ms.halfScale);
      }
      for (int i = (height>>1)-inc; i > ms.scale; i -= inc) {
        line(0, i, width, i);
        String mapI = map(i, 0, height, f.yMin+f.wh, f.yMin)+"i";
        text(mapI, (width-int(textWidth(mapI)))>>1, i-ms.halfScale);
      }
    }

    int currWidth = int(textWidth(state))+halfScale;
    int currLeft = max(min(x, width-currWidth-halfScale), halfScale);

    fill(0xc0202020);
    stroke(0xffe0e0e0);
    rect(currLeft, height-doubleScale, currWidth, oneAndHalfScale);
    fill(0xffe0e0e0);
    text(state, currLeft+quarterScale, height-doubleScale+quarterScale);
  }
}

class Menu {
  boolean visible;
  int w, h;
  String title;
  Menu parent;
  ArrayList<Menu> children;

  Menu(String inTitle, Menu inParent) {
    h = 0;
    title = inTitle;
    parent = inParent;
    children = new ArrayList<Menu>();
  }

  Menu addMenu(String inTitle) {
    children.add(new Menu(inTitle, this));

    int currW = int(textWidth(inTitle))+ms.halfScale;
    w = (currW > w) ? currW : w;
    h += ms.oneAndHalfScale;

    return children.get(children.size()-1);
  }

  void correctPosition(int inW, int inH) {
    for (int i = 0; i < children.size(); i++) {
      ms.x = max(min(ms.x, width-inW-w-ms.halfScale), ms.halfScale);
      ms.y = max(min(ms.y, height-inH-h-ms.doubleScale-ms.halfScale), ms.halfScale);

      Menu iChild = children.get(i);
      if (iChild.visible) iChild.correctPosition(inW+w, inH+i*ms.oneAndHalfScale);
    }
  }

  void display(int inX, int inY) {
    if (visible) {
      fill(0xc0202020);
      stroke(0xffe0e0e0);
      rect(inX, inY, w, h);
      for (int i = 0; i < children.size(); i++) {
        int currTop = inY+i*ms.oneAndHalfScale;

        Menu iChild = children.get(i);
        fill(0xffe0e0e0);
        text(iChild.title, inX+ms.quarterScale, currTop+ms.quarterScale);
        if (iChild.children.size() > 0) iChild.display(inX+w, currTop);
      }
    }
  }

  void doAction() {
    if (title.equals("Close")) {
      ms.mode = "";
      r.hideAll();
      wc.hide();
    }
    else if (title.equals("Display Complex Grid")) {
      ms.showGrid = !ms.showGrid;
      ms.mode = (ms.showGrid) ? "Display Complex Grid" : "";
    }
    else if (title.equals("Display Orbit Trap")) {
      if (ms.mode.equals("Display Orbit Trap")) ms.mode = "";
      else ms.mode = "Display Orbit Trap";
    }
    else if (title.equals("Fractal Color")) {
      ms.mode = "Fractal Color";
      ms.showGrid = false;
      wc.visible = true;
      wc.correctPosition();
    }
    else if (title.equals("Julia / Mandelbrot Mode")) {
      ms.mode = "Julia / Mandelbrot Mode";
      f.juliaMode = !f.juliaMode;

      Menu currMenu = r.getMenu("Fractal Options");
      if (f.juliaMode) {
        currMenu.addMenu("Set Constant");
        currMenu = currMenu.addMenu("Suggested Constants");
        for (int i = 0; i < f.suggestions.size(); i++) {
          currMenu.addMenu(f.suggestions.get(i).r+", "+f.suggestions.get(i).i+"i");
        }
        r.correctPosition(0, 0);
      }
      else {
        currMenu.removeMenu("Set Constant");
        currMenu.removeMenu("Suggested Constants");
      }

      wc.shaderUpdateNodes();
      f.reset();
      f.shaderRender();
    }
    else if (title.equals("Pan")) {
      if (ms.mode.equals("Pan")) ms.mode = "";
      else ms.mode = "Pan";
    }
    else if (title.equals("Reset")) {
      ms.mode = "Reset";
      f.reset();
      f.shaderRender();
    }
    else if (title.equals("Set Constant")) {
      if (ms.mode.equals("Set Constant")) {
        ms.mode = "";
        ms.showGrid = false;
      }
      else {
        ms.mode = "Set Constant";
        ms.showGrid = true;
        f.reset();
        f.shaderRender();
      }
    }
    else if (title.equals("Zoom In")) {
      if (ms.mode.equals("Zoom In")) ms.mode = "";
      else ms.mode = "Zoom In";
    }
    else if (title.equals("Zoom Out")) {
      if (ms.mode.equals("Zoom Out")) ms.mode = "";
      else ms.mode = "Zoom Out";
    }
    else if (this.parent.title.equals("Suggested Constants")) {
      ms.mode = "Julia / Mandelbrot Mode";

      Menu currMenu = r.getMenu("Suggested Constants");
      for (int i = 0; i < currMenu.children.size(); i++) {
        Menu iChild = currMenu.children.get(i);
        if (iChild.title.equals(title)) {
          f.jSelected = i;
          break;
        }
      }
      f.jr = f.suggestions.get(f.jSelected).r;
      f.ji = f.suggestions.get(f.jSelected).i;
      f.jSet.set("jr", f.jr);
      f.jSet.set("ji", f.ji);

      f.shaderRender();
    }
    else wc.hide();
  }

  Menu getMenu(String inTitle) {
    Menu out = null;

    for (int i = 0; i < children.size(); i++) {
      Menu iChild = children.get(i);
      if (iChild.title.equals(inTitle)) {
        out = iChild;
        break;
      }
      out = iChild.getMenu(inTitle);
      if (out != null) break;
    }

    return out;
  }

  void hideAll() {
    visible = false;
    for (int i = 0; i < children.size(); i++) children.get(i).hideAll();
  }

  Menu pointInMenu(int inMx, int inMy, int inPx, int inPy) {
    Menu out = null;

    for (int i = 0; i < children.size(); i++) {
      int currTop = inMy+i*ms.oneAndHalfScale;
      int currBottom = inMy+(i+1)*ms.oneAndHalfScale;

      Menu iChild = children.get(i);
      if (inPx >= inMx && inPx < inMx+w && inPy >= currTop && inPy < currBottom) {
        out = iChild;
        break;
      }
      else if (iChild.visible) {
        out = iChild.pointInMenu(inMx+w, currTop, inPx, inPy);
        if (out != null) break;
      }
    }

    return out;
  }

  void showMenuPath() {
    this.visible = true;
    if (this.parent != null) this.parent.showMenuPath();
  }

  void removeMenu(String inTitle) {
    for (int i = 0; i < children.size(); i++) {
      Menu iChild = children.get(i);
      if (iChild.title.equals(inTitle)) {
        children.remove(i);
        break;
      }
    }

    w = 0;
    for (int i = 0; i < children.size(); i++) {
      int currWidth = int(textWidth(children.get(i).title))+ms.halfScale;
      w = (currWidth > w) ? currWidth : w;
    }
    h -= ms.oneAndHalfScale;
  }
}

class Fractal {
  boolean juliaMode;
  int jSelected, maxIterations, oIterations;
  float jr, ji, oi, or, wh, xMin, yMin;
  ArrayList<ComplexN> suggestions;
  PGraphics pg;
  PShader jSet, mSet;

  Fractal() {
    juliaMode = false;
    jSelected = 0;

    suggestions = new ArrayList<ComplexN>();
    suggestions.add(new ComplexN(-0.8, 0.156));
    suggestions.add(new ComplexN(-0.74543, 0.11301));
    suggestions.add(new ComplexN(-0.70176, 0.3842));
    suggestions.add(new ComplexN(0.233, 0.5378));
    suggestions.add(new ComplexN(0.285, 0.01));

    jr = suggestions.get(0).r;
    ji = suggestions.get(0).i;

    pg = createGraphics(width, height, P2D);

    jSet = loadShader("jSet.glsl");
    jSet.set("pixelDimensions", float(pg.width), float(pg.height));
    jSet.set("processingOffset", float(ms.halfScale));
    jSet.set("jr", jr);
    jSet.set("ji", ji);

    mSet = loadShader("mSet.glsl");
    mSet.set("pixelDimensions", float(pg.width), float(pg.height));
    mSet.set("processingOffset", float(ms.halfScale));

    reset();
  }

  class ComplexN {
    float r, i;
    ComplexN(float inR, float inI) {
      r = inR;
      i = inI;
    }
  }

  void displayOrbitTrap() {
    int maxPwh = max(width, height);
    float delta = f.wh/maxPwh*0.5;
    float xOffset = min((width-height)*delta, 0.0);
    float yOffset = min((height-width)*delta, 0.0);

    float mXMin = xMin-xOffset;
    float mXMax = xMin+wh+xOffset;
    float mYMin = yMin+wh+yOffset;
    float mYMax = yMin-yOffset;
    PVector p = new PVector(map(or, mXMin, mXMax, 0.0, width), map(oi, mYMin, mYMax, 0.0, height));

    float zr = or;
    float zi = oi;
    oIterations = 0;
    if (juliaMode) {
      for (; oIterations < maxIterations; oIterations++) {
        float sqZr = zr*zr;
        float sqZi = zi*zi;
        float twoZri = 2.0*zr*zi;
        zr = sqZr-sqZi+jr;
        zi = twoZri+ji;
        if (sqZr+sqZi > 16.0) break;

        float mapZr = map(zr, mXMin, mXMax, 0.0, width);
        float mapZi = map(zi, mYMin, mYMax, 0.0, height);
        stroke(0x40202020);
        strokeWeight(3);
        line(p.x, p.y, mapZr, mapZi);
        stroke(0xc0e0e0e0);
        strokeWeight(1);
        line(p.x, p.y, mapZr, mapZi);
        p.x = mapZr;
        p.y = mapZi;
      }
    }
    else {
      for (; oIterations < maxIterations; oIterations++) {
        float sqZr = zr*zr;
        float sqZi = zi*zi;
        float twoZri = 2.0*zr*zi;
        zr = sqZr-sqZi+or;
        zi = twoZri+oi;
        if (sqZr+sqZi > 16.0) break;

        float mapZr = map(zr, mXMin, mXMax, 0.0, width);
        float mapZi = map(zi, mYMin, mYMax, 0.0, height);
        stroke(0x40202020);
        strokeWeight(3);
        line(p.x, p.y, mapZr, mapZi);
        stroke(0xc0e0e0e0);
        strokeWeight(1);
        line(p.x, p.y, mapZr, mapZi);
        p.x = mapZr;
        p.y = mapZi;
      }
    }
  }

  void reset() {
    oi = 0.0;
    or = 0.0;

    maxIterations = 1000;
    wh = 4.0;
    xMin = (juliaMode) ? -2.0 : -2.5;
    yMin = -2.0;

    if (juliaMode) {
      jSet.set("maxIterations", maxIterations);
      jSet.set("wh", wh);
      jSet.set("xMin", xMin);
      jSet.set("yMin", yMin);
    }
    else {
      mSet.set("maxIterations", maxIterations);
      mSet.set("wh", wh);
      mSet.set("xMin", xMin);
      mSet.set("yMin", yMin);
    }
  }

  void shaderRender() {
    pg.beginDraw();
    if (juliaMode) pg.shader(jSet);
    else pg.shader(mSet);
    pg.rect(0, 0, pg.width, pg.height);
    pg.endDraw();
  }
}

class WidgetColor {
  boolean visible;
  int selected;
  String mode;
  String[] options;
  ArrayList<Node> nodes;
  PGraphics pg;
  PShader ps;

  WidgetColor() {
    visible = false;
    selected = -1;
    mode = "";

    options = new String[4];
    options[0] = "Color";
    options[1] = "Move";
    options[2] = "Add";
    options[3] = "Remove";

    pg = createGraphics(width-ms.scale, ms.halfScale, P2D);
    ps = loadShader("widgetColor.glsl");
    ps.set("processingOffset", float(ms.halfScale));

    nodes = new ArrayList<Node>();
    nodes.add(new Node(ms.halfScale, 0xff000000));
    nodes.add(new Node(width-ms.scale, 0xffffffff));
    shaderUpdateNodes();
  }

  class Node {
    boolean updating;
    int c, x;

    Node(int inX, int inColor) {
      updating = false;
      x = inX;
      c = inColor;
    }

    void display(int inY) {
      fill(c);
      stroke(0xffe0e0e0);
      rect(x, inY-ms.scale, ms.halfScale, ms.halfScale);

      if (updating) {
        int startLeft = x-(ms.scale<<2);
        startLeft = max(min(startLeft, width-(ms.scale<<3)-ms.oneAndHalfScale)+ms.quarterScale, ms.halfScale);
        int startTop = inY-ms.doubleScale;

        for (int i = 0; i < 3; i++) {
          int currTop = startTop-i*ms.scale;
          fill(0xc0202020);
          rect(startLeft+ms.halfScale, currTop, ms.scale<<3, ms.halfScale);

          int iShift = i<<3;
          int channelVal = c&(0xff<<iShift);
          fill(0xff000000|channelVal);
          rect(startLeft, currTop, ms.halfScale, ms.halfScale);

          fill(0xffe0e0e0);
          int mx = startLeft+ms.halfScale+int(map(channelVal>>iShift, 0, 255, 0.0, ms.scale<<3));
          rect(mx, currTop, ms.quarterScale, ms.halfScale);
        }
      }
    }
  }

  void addNode(int inX, int inColor) {
    nodes.add(new Node(inX, inColor));
    sortNodes();
    shaderUpdateNodes();
    f.shaderRender();
  }

  boolean conflictingNodeMenu() {
    boolean out = false;

    int cWidth = (ms.scale<<3)+ms.halfScale;
    for (int i = 0; i < nodes.size(); i++) {
      Node iNode = nodes.get(i);
      if (!iNode.updating) continue;

      int currLeft = iNode.x-(ms.scale<<2);
      currLeft = max(min(currLeft, width-cWidth-ms.oneAndHalfScale)+ms.quarterScale, ms.halfScale);
      if (ms.x < currLeft+cWidth && ms.x+r.w > currLeft) {
        out = true;
        break;
      }
    }

    return out;
  }

  void conflictingNodeUpdates() {
    int cWidth = (ms.scale<<3)+ms.halfScale;
    int selectedLeft = nodes.get(selected).x-(ms.scale<<2);
    selectedLeft = max(min(selectedLeft, width-cWidth-ms.oneAndHalfScale)+ms.quarterScale, ms.halfScale);

    for (int i = 0; i < nodes.size(); i++) {
      if (i == selected) continue;

      Node iNode = nodes.get(i);
      int currLeft = iNode.x-(ms.scale<<2);
      currLeft = max(min(currLeft, width-cWidth-ms.oneAndHalfScale)+ms.quarterScale, ms.halfScale);
      if (selectedLeft < currLeft+cWidth && selectedLeft+cWidth > currLeft) iNode.updating = false;
    }
  }

  void correctPosition() {
    int totalWidth = ms.doubleScale;
    for (int i = 0; i < 4; i++) totalWidth += int(textWidth(options[i]))+ms.halfScale;

    ms.x = min(ms.x, width-totalWidth);
    ms.y = min(ms.y, height-r.h-(ms.scale*6)-ms.halfScale);
  }

  void display() {
    int currTop = ms.y+r.h+ms.oneAndHalfScale;
    for (int i = 0; i < nodes.size(); i++) nodes.get(i).display(currTop);

    stroke(0xffe0e0e0);
    rect(ms.halfScale, currTop, pg.width, pg.height);
    image(pg, ms.halfScale, currTop, pg.width, pg.height);

    int currLeft = ms.x;
    for (int i = 0; i < 4; i++) {
      int currWidth = int(textWidth(options[i]))+ms.halfScale;
      fill(0xc0202020);
      rect(currLeft, currTop+ms.scale, currWidth, ms.oneAndHalfScale);
      fill(0xffe0e0e0);
      text(options[i], currLeft+ms.quarterScale, currTop+ms.scale+ms.quarterScale);
      currLeft += currWidth+ms.halfScale;
    }
  }

  void hide() {
    visible = false;
    selected = -1;
    mode = "";
    hideUpdate();
  }

  void hideUpdate() {
    for (int i = 0; i < nodes.size(); i++) nodes.get(i).updating = false;
  }

  void pointInNode(int inX, int inY) {
    selected = -1;

    int currTop = ms.y+r.h+ms.halfScale;
    for (int i = 0; i < nodes.size(); i++) {
      Node iNode = nodes.get(i);
      if (inX >= iNode.x && inX < iNode.x+ms.halfScale && inY >= currTop && inY < currTop+ms.halfScale) {
        selected = i;
        break;
      }
    }
  }

  void pointInOption(int inX, int inY) {
    mode = "";

    int currLeft = ms.x;
    int currTop = ms.y+r.h+ms.oneAndHalfScale+ms.scale;
    int currBottom = currTop+ms.oneAndHalfScale+ms.quarterScale;
    for (int i = 0; i < 4; i++) {
      int currWidth = int(textWidth(options[i]))+ms.halfScale;
      if (inX >= currLeft && inX < currLeft+currWidth && inY >= currTop && inY < currBottom) {
        mode = options[i];
        break;
      }
      currLeft += currWidth+ms.halfScale;
    }
  }

  void removeNode(int inIndex) {
    nodes.remove(inIndex);
    sortNodes();
    shaderUpdateNodes();
    f.shaderRender();
  }

  void shaderRender() {
    pg.beginDraw();
    pg.shader(ps);
    pg.rect(0, 0, pg.width, pg.height);
    pg.endDraw();
  }

  void shaderUpdateNodes() {
    int nodeSize = nodes.size();

    float[] nodeX = new float[nodeSize];
    float[] nodeR = new float[nodeSize];
    float[] nodeG = new float[nodeSize];
    float[] nodeB = new float[nodeSize];

    float colorDiv = 1.0/255.0;
    for (int i = 0; i < nodeSize; i++) {
      nodeX[i] = float(nodes.get(i).x);
      nodeR[i] = ((nodes.get(i).c&0xff0000)>>16)*colorDiv;
      nodeG[i] = ((nodes.get(i).c&0xff00)>>8)*colorDiv;
      nodeB[i] = (nodes.get(i).c&0xff)*colorDiv;
    }

    ps.set("numNodes", nodeSize);
    ps.set("nodeX", nodeX);
    ps.set("nodeR", nodeR);
    ps.set("nodeG", nodeG);
    ps.set("nodeB", nodeB);
    shaderRender();

    if (f.juliaMode) {
      f.jSet.set("numNodes", nodeSize);
      f.jSet.set("nodeX", nodeX);
      f.jSet.set("nodeR", nodeR);
      f.jSet.set("nodeG", nodeG);
      f.jSet.set("nodeB", nodeB);
    }
    else {
      f.mSet.set("numNodes", nodeSize);
      f.mSet.set("nodeX", nodeX);
      f.mSet.set("nodeR", nodeR);
      f.mSet.set("nodeG", nodeG);
      f.mSet.set("nodeB", nodeB);
    }
  }

  void sortNodes() {
    boolean swapped = true;
    int endIndex = 1;
    while (swapped) {
      swapped = false;
      for (int i = 0; i < nodes.size()-endIndex; i++) {
        Node leftNode = nodes.get(i);
        Node rightNode = nodes.get(i+1);
        if (leftNode.x > rightNode.x) {
          boolean swapU = leftNode.updating;
          int swapX = leftNode.x;
          int swapC = leftNode.c;
          leftNode.updating = rightNode.updating;
          leftNode.x = rightNode.x;
          leftNode.c = rightNode.c;
          rightNode.updating = swapU;
          rightNode.x = swapX;
          rightNode.c = swapC;
          swapped = true;
        }
      }
      endIndex++;
    }
  }
}

