#define PROCESSING_COLOR_SHADER

const float infinity = 1.0/0.0;

uniform int numNodes;
uniform float processingOffset;
uniform float nodeX[32];
uniform float nodeR[32];
uniform float nodeG[32];
uniform float nodeB[32];

void main(void) {
	// Assign offsetGLCoord based on gl_FragCoord and the offset in Processing
	float offsetGlCoord = gl_FragCoord.x+processingOffset;
	
	// Determine the left and right node indices relative to offsetGlCoord
	int leftIndex = 0;
	int rightIndex = numNodes-1;
	float leftClosest = infinity;
	float rightClosest = infinity;
	for (int i = 0; i < numNodes; i++) {
		float leftDist = offsetGlCoord-nodeX[i];
		float rightDist = nodeX[i]-offsetGlCoord;
		if (leftDist > 0.0 && leftDist < leftClosest) {
			leftIndex = i;
			leftClosest = leftDist;
		}
		if (rightDist > 0.0 && rightDist < rightClosest) {
			rightIndex = i;
			rightClosest = rightDist;
		}
	}
	
	// Interpolate unless offsetGlCoord < all x, offsetGlCoord > all x, or there was a tie
	if (leftIndex != rightIndex) {
		// Assign parametric t values
		float t = (offsetGlCoord-nodeX[leftIndex])/(nodeX[rightIndex]-nodeX[leftIndex]);
		float it = 1.0-t;
		
		// Interpolate among left and right node indices
		gl_FragColor = vec4(
			((nodeR[leftIndex]*it)+(nodeR[rightIndex]*t)), 
			((nodeG[leftIndex]*it)+(nodeG[rightIndex]*t)), 
			((nodeB[leftIndex]*it)+(nodeB[rightIndex]*t)), 
		1.0);
	}
	else gl_FragColor = vec4(nodeR[leftIndex], nodeG[leftIndex], nodeB[leftIndex], 1.0);
}
